#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import decimal

from tests import event_loop
from tests.exchanges import simulated_exchange_manager, simulated_trader
from tests.personal_data.trades import create_executed_trade

import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.constants as constants


SYMBOL = "BTC/USDT"


def test_empty():
    pnl = personal_data.TradePnl([], [])
    with pytest.raises(errors.IncompletePNLError):
        pnl.get_entry_time()
    with pytest.raises(errors.IncompletePNLError):
        pnl.get_close_time()
    assert pnl.get_total_entry_quantity() == constants.ZERO
    assert pnl.get_total_close_quantity() == constants.ZERO
    with pytest.raises(errors.IncompletePNLError):
        pnl.get_entry_price()
    with pytest.raises(errors.IncompletePNLError):
        pnl.get_close_price()
    assert pnl.get_paid_regular_fees_in_quote() == constants.ZERO
    assert pnl.get_paid_special_fees_by_currency() == {}


def test_with_invalid_values(simulated_trader):
    _, _, trader = simulated_trader
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(0), decimal.Decimal(10), SYMBOL, _get_fees("BTC", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 15, decimal.Decimal(1), decimal.Decimal(15), SYMBOL, _get_fees("USDT", 1)),
    ])
    with pytest.raises(errors.IncompletePNLError):
        pnl.get_close_ratio()
    with pytest.raises(errors.IncompletePNLError):
        pnl.get_profits()
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(1), decimal.Decimal(0), SYMBOL, _get_fees("BTC", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 15, decimal.Decimal(1), decimal.Decimal(15), SYMBOL, _get_fees("USDT", 1)),
    ])
    assert pnl.get_close_ratio() == constants.ONE
    assert pnl.get_profits() == (
        decimal.Decimal("14"),  # 15 - fees
        constants.ZERO
    )


def test_simple_buy_entry_and_sell_close(simulated_trader):
    _, _, trader = simulated_trader
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(1), decimal.Decimal(10), SYMBOL, _get_fees("BTC", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 15, decimal.Decimal(1), decimal.Decimal(15), SYMBOL, _get_fees("USDT", 1)),
    ])
    assert pnl.get_entry_time() == 12
    assert pnl.get_close_time() == 15
    assert pnl.get_total_entry_quantity() == decimal.Decimal(1)
    assert pnl.get_total_close_quantity() == decimal.Decimal(1)
    assert pnl.get_entry_price() == decimal.Decimal(10)
    assert pnl.get_close_price() == decimal.Decimal(15)
    assert pnl.get_closed_entry_value() == decimal.Decimal(10) * decimal.Decimal(1)
    assert pnl.get_closed_close_value() == decimal.Decimal(15) * decimal.Decimal(1)
    assert pnl.get_close_ratio() == decimal.Decimal("1")
    assert pnl.get_closed_pnl_quantity() == decimal.Decimal(1)
    assert pnl.get_paid_regular_fees_in_quote() == decimal.Decimal(1) + decimal.Decimal("0.1") * decimal.Decimal(10)
    assert pnl.get_paid_special_fees_by_currency() == {}
    # started with 10, closed with 15, paid 2 in fees: end value: 13
    assert pnl.get_profits() == (
        decimal.Decimal(15) - decimal.Decimal(10) - decimal.Decimal(2),  # close - entry - fees
        decimal.Decimal(30)  # 10 + 30% = 13
    )


def test_simple_sell_entry_and_buy_close(simulated_trader):
    _, _, trader = simulated_trader
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 12, decimal.Decimal(1), decimal.Decimal(15), SYMBOL, _get_fees("BTC", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 15, decimal.Decimal("1.5"), decimal.Decimal(10), SYMBOL, _get_fees("USDT", 1)),
    ])
    assert pnl.get_entry_time() == 12
    assert pnl.get_close_time() == 15
    assert pnl.get_total_entry_quantity() == decimal.Decimal(1)
    assert pnl.get_total_close_quantity() == decimal.Decimal("1.5")
    assert pnl.get_entry_price() == decimal.Decimal(15)
    assert pnl.get_close_price() == decimal.Decimal(10)
    assert pnl.get_closed_entry_value() == decimal.Decimal(15) * decimal.Decimal(1)
    assert pnl.get_closed_close_value() == decimal.Decimal(10) * decimal.Decimal("1.5")
    assert pnl.get_close_ratio() == decimal.Decimal("1")
    assert pnl.get_closed_pnl_quantity() == decimal.Decimal("1.5")
    assert pnl.get_paid_regular_fees_in_quote() == decimal.Decimal(1) + decimal.Decimal("0.1") * decimal.Decimal(15)
    assert pnl.get_paid_special_fees_by_currency() == {}
    # started with 15, closed with 15, paid 2.5 in fees: end value: 12.5
    assert pnl.get_profits() == (
        decimal.Decimal(15) - decimal.Decimal(15) - decimal.Decimal("2.5"),  # close - entry - fees
        decimal.Decimal("-16.66666666666666666666666667")  # 15 - 16.67% = 12.5
    )


def test_simple_sell_entry_and_buy_close_with_more_funds(simulated_trader):
    _, _, trader = simulated_trader
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 12, decimal.Decimal(1), decimal.Decimal(15), SYMBOL, _get_fees("BTC", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 15, decimal.Decimal(2), decimal.Decimal(10), SYMBOL, _get_fees("USDT", 1)),
    ])
    assert pnl.get_entry_time() == 12
    assert pnl.get_close_time() == 15
    assert pnl.get_total_entry_quantity() == decimal.Decimal(1)
    assert pnl.get_total_close_quantity() == decimal.Decimal(2)
    assert pnl.get_entry_price() == decimal.Decimal(15)
    assert pnl.get_close_price() == decimal.Decimal(10)
    assert pnl.get_closed_entry_value() == decimal.Decimal(15) * decimal.Decimal(1)
    # use "1.5" as closed quantity for value and PNL as only 1.5 come from this entry
    assert pnl.get_closed_close_value() == decimal.Decimal(10) * decimal.Decimal("1.5")
    assert pnl.get_close_ratio() == decimal.Decimal("1")
    assert pnl.get_closed_pnl_quantity() == decimal.Decimal("1.5")
    assert pnl.get_paid_regular_fees_in_quote() == decimal.Decimal(1) + decimal.Decimal("0.1") * decimal.Decimal(15)
    assert pnl.get_paid_special_fees_by_currency() == {}
    # started with 15, closed with 20 (-5 to align with entry), paid 2.5 in fees: end value: 12.5
    assert pnl.get_profits() == (
        decimal.Decimal(15) - decimal.Decimal(15) - decimal.Decimal("2.5"),  # close - entry - fees
        decimal.Decimal("-16.66666666666666666666666667")  # 15 - 16.67% = 12.5
    )


def test_simple_sell_entry_and_buy_close_with_less_funds(simulated_trader):
    _, _, trader = simulated_trader
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 12, decimal.Decimal(1), decimal.Decimal(15), SYMBOL, _get_fees("BTC", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 15, decimal.Decimal(1), decimal.Decimal(10), SYMBOL, _get_fees("USDT", 1)),
    ])
    assert pnl.get_entry_time() == 12
    assert pnl.get_close_time() == 15
    assert pnl.get_total_entry_quantity() == decimal.Decimal(1)
    assert pnl.get_total_close_quantity() == decimal.Decimal(1)
    assert pnl.get_entry_price() == decimal.Decimal(15)
    assert pnl.get_close_price() == decimal.Decimal(10)
    assert pnl.get_closed_entry_value() == decimal.Decimal(15) * decimal.Decimal(1)
    assert pnl.get_closed_close_value() == decimal.Decimal(10) * decimal.Decimal(1)
    assert pnl.get_close_ratio() == decimal.Decimal("0.6666666666666666666666666667")
    assert pnl.get_closed_pnl_quantity() == decimal.Decimal(1)
    assert pnl.get_paid_regular_fees_in_quote() == decimal.Decimal(1) + decimal.Decimal("0.1") * decimal.Decimal(15)
    # started with 15(-5 because of close cost), closed with 10, paid 2.5 in fees: end value: 7.5
    assert pnl.get_profits() == (
        decimal.Decimal(10) - decimal.Decimal(10) - decimal.Decimal("2.5"),  # close - entry - fees
        decimal.Decimal("-25")  # 10 - 25% = 12.5
    )


def test_simple_entry_and_double_close(simulated_trader):
    _, _, trader = simulated_trader
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(1), decimal.Decimal(10), SYMBOL, _get_fees("BTC", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 15, decimal.Decimal("0.4"), decimal.Decimal(15), SYMBOL, _get_fees("USDT", 0.5)),
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 16, decimal.Decimal(1), decimal.Decimal(16), SYMBOL, _get_fees("USDT", 1)),
    ])
    assert pnl.get_entry_time() == 12
    assert pnl.get_close_time() == 16
    assert pnl.get_total_entry_quantity() == decimal.Decimal(1)
    assert pnl.get_total_close_quantity() == decimal.Decimal("1.4")
    assert pnl.get_entry_price() == decimal.Decimal(10)
    assert pnl.get_close_price() == decimal.Decimal("15.5")
    assert pnl.get_closed_entry_value() == decimal.Decimal(10) * decimal.Decimal(1)  # 1 as entry is 1
    assert pnl.get_closed_close_value() == decimal.Decimal("15.5") * decimal.Decimal(1)  # 1 as entry is 1
    assert pnl.get_close_ratio() == decimal.Decimal(1)
    assert pnl.get_closed_pnl_quantity() == decimal.Decimal(1)  # 1 as entry is 1
    assert pnl.get_paid_regular_fees_in_quote() == decimal.Decimal(1) + decimal.Decimal("0.5") + \
           decimal.Decimal("0.1") * decimal.Decimal(10)
    # started with 10, closed with 15.5, paid 2.5 in fees: end value: 13
    assert pnl.get_profits() == (
        decimal.Decimal(15) - decimal.Decimal(10) - decimal.Decimal(2),  # close - entry - fees
        decimal.Decimal(30)  # 10 + 30% = 13
    )


def test_double_simple_entry_and_double_close(simulated_trader):
    _, _, trader = simulated_trader
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(1), decimal.Decimal(10), SYMBOL, _get_fees("BTC", 0.1)),
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 9, decimal.Decimal("0.5"), decimal.Decimal(9), SYMBOL, _get_fees("BTC", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 15, decimal.Decimal("0.4"), decimal.Decimal(15), SYMBOL, _get_fees("USDT", 0.5)),
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 16, decimal.Decimal(1), decimal.Decimal(16), SYMBOL, _get_fees("USDT", 1)),
    ])
    assert pnl.get_entry_time() == 9
    assert pnl.get_close_time() == 16
    assert pnl.get_total_entry_quantity() == decimal.Decimal("1.5")
    assert pnl.get_total_close_quantity() == decimal.Decimal("1.4")
    assert pnl.get_entry_price() == decimal.Decimal("9.5")
    assert pnl.get_close_price() == decimal.Decimal("15.5")
    assert pnl.get_closed_entry_value() == decimal.Decimal("9.5") * decimal.Decimal("1.5")
    assert pnl.get_closed_close_value() == decimal.Decimal("15.5") * decimal.Decimal("1.4")  # 1.4 as close is 1.4
    assert pnl.get_close_ratio() == decimal.Decimal("1.4") / decimal.Decimal("1.5")
    assert pnl.get_closed_pnl_quantity() == decimal.Decimal("1.4")  # 1.4 as close is 1.4
    assert pnl.get_paid_regular_fees_in_quote() == decimal.Decimal(1) + decimal.Decimal("0.5") + \
           decimal.Decimal("0.1") * decimal.Decimal(10) + decimal.Decimal("0.1") * decimal.Decimal(9)
    assert pnl.get_paid_special_fees_by_currency() == {}
    # started with 1.4*9.5=13.3, closed with 1.4*15.5=21.7, paid 3.4 in fees: end value: 5.1
    assert pnl.get_profits() == (
        decimal.Decimal("5"),
        decimal.Decimal("37.5939849624060150375939850")
    )


def test_get_paid_special_fees_by_currency(simulated_trader):
    _, _, trader = simulated_trader
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(1), decimal.Decimal(10), SYMBOL, _get_fees("BNB", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 15, decimal.Decimal("0.4"), decimal.Decimal(15), SYMBOL, _get_fees("USDT", 0.5)),
    ])
    assert pnl.get_paid_special_fees_by_currency() == {"BNB": decimal.Decimal(str(0.1))}
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(1), decimal.Decimal(10), SYMBOL, _get_fees("BNB", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 15, decimal.Decimal("0.4"), decimal.Decimal(15), SYMBOL, _get_fees("PLOP", 0.5)),
    ])
    assert pnl.get_paid_special_fees_by_currency() == {"BNB": decimal.Decimal(str(0.1)), "PLOP": decimal.Decimal(str(0.5))}
    pnl = personal_data.TradePnl([
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(1), decimal.Decimal(10), SYMBOL, _get_fees("BNB", 0.1)),
        create_executed_trade(trader, enums.TradeOrderSide.BUY, 12, decimal.Decimal(1), decimal.Decimal(10), SYMBOL, _get_fees("PLOP", 0.1)),
    ], [
        create_executed_trade(trader, enums.TradeOrderSide.SELL, 15, decimal.Decimal("0.4"), decimal.Decimal(15), SYMBOL, _get_fees("PLOP", 0.5)),
    ])
    assert pnl.get_paid_special_fees_by_currency() == {"BNB": decimal.Decimal(str(0.1)), "PLOP": decimal.Decimal(str(0.6))}


def _get_fees(currency, value):
    return {
        enums.FeePropertyColumns.CURRENCY.value: currency,
        enums.FeePropertyColumns.COST.value: decimal.Decimal(str(value))
    }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal

import octobot_commons.logging as logging
import octobot_commons.symbols as symbol_util
import octobot_commons.asyncio_tools as asyncio_tools

import octobot_trading.constants as constants
import octobot_trading.errors as errors


class ValueConverter:
    """
    ValueConverter manipulates trading pairs values to extract asset values
    Uses a portfolio_manager and portfolio_value_holder to read trading pairs values
    """
    MAX_PRICE_BRIDGE_DEPTH = 6

    def __init__(self, portfolio_manager):
        self.portfolio_manager = portfolio_manager
        self._bot_main_loop = asyncio.get_event_loop()
        self.logger = logging.get_logger(f"{self.__class__.__name__}"
                                         f"[{self.portfolio_manager.exchange_manager.exchange_name}]")

        self.last_prices_by_trading_pair = {}

        self.initializing_symbol_prices = set()
        self.initializing_symbol_prices_pairs = set()

        # set of currencies for which the current exchange is not providing any suitable price data
        self.missing_currency_data_in_exchange = set()

        # internal price conversion elements
        self._price_bridge_by_symbol = {}
        self._missing_price_bridges = set()

    def update_last_price(self, symbol, price):
        if symbol not in self.last_prices_by_trading_pair:
            self.reset_missing_price_bridges()
            self.logger.debug(f"Initialized last price for {symbol}")
        self.last_prices_by_trading_pair[symbol] = price

    def evaluate_value(self, currency, quantity, raise_error=True):
        """
        Evaluate value returns the currency quantity value in the reference (attribute) currency
        :param currency: the currency to evaluate
        :param quantity: the currency quantity
        :param raise_error: will catch exception if False
        :return: the currency value
        """
        # easy case --> the current currency is the reference currency or the quantity is 0
        if currency == self.portfolio_manager.reference_market or quantity == constants.ZERO:
            return quantity
        currency_value = self._try_get_value_of_currency(currency, quantity, raise_error)
        return self._check_currency_initialization(currency, currency_value)

    def _check_currency_initialization(self, currency, currency_value):
        """
        Check if the currency has to be removed from self.initializing_symbol_prices and return currency_value
        :param currency: the currency to check
        :param currency_value: the currency value
        :return: currency_value after checking
        """
        if currency_value > constants.ZERO and currency in self.initializing_symbol_prices:
            self.initializing_symbol_prices.remove(currency)
        return currency_value

    def _try_get_value_of_currency(self, currency, quantity, raise_error):
        """
        try_get_value_of_currency will try to get the value of the given currency quantity in reference market.
        It will try to get it from a trading pair that fit with the exchange availability.
        :return: the value found of this currency quantity, if not found returns 0.
        """
        settlement_asset = self.portfolio_manager.reference_market \
            if self.portfolio_manager.exchange_manager.is_future else None
        try:
            # 1. try from existing pairs (as is and reversed)
            return self.convert_currency_value_using_last_prices(
                quantity, currency,
                self.portfolio_manager.reference_market,
                settlement_asset=settlement_asset
            )
        except errors.MissingPriceDataError as missing_data_exception:
            if not self.portfolio_manager.exchange_manager.is_future:
                try:
                    # 2. try from existing indirect pairs
                    value = self.try_convert_currency_value_using_multiple_pairs(
                        currency, self.portfolio_manager.reference_market, quantity, []
                    )
                    if value is not None:
                        return value
                except (errors.MissingPriceDataError, errors.PendingPriceDataError):
                    pass
            symbol = symbol_util.merge_currencies(
                currency, self.portfolio_manager.reference_market, settlement_asset=settlement_asset
            )
            reversed_symbol = symbol_util.merge_currencies(
                self.portfolio_manager.reference_market, currency, settlement_asset=settlement_asset
            )
            # 3. if the pair or reversed pair is traded on exchange, use it to price "currency"
            if not any(
                self.portfolio_manager.exchange_manager.symbol_exists(s)
                for s in (symbol, reversed_symbol)
            ) and currency not in self.missing_currency_data_in_exchange:
                self._inform_no_matching_symbol(currency)
                self.missing_currency_data_in_exchange.add(currency)
            if not self.portfolio_manager.exchange_manager.is_backtesting:
                self._try_to_ask_ticker_missing_symbol_data(currency, symbol, reversed_symbol)
                if raise_error:
                    raise missing_data_exception
        return constants.ZERO

    def _try_to_ask_ticker_missing_symbol_data(self, currency, symbol, reversed_symbol):
        """
        Try to ask the ticker producer to watch additional symbols
        to collect missing data required for profitability calculation
        :param currency: the concerned currency
        :param symbol: the symbol to add
        :param reversed_symbol: the reversed symbol to add
        """
        symbols_to_add = []
        if self.portfolio_manager.exchange_manager.symbol_exists(symbol):
            symbols_to_add = [symbol]
        elif self.portfolio_manager.exchange_manager.symbol_exists(reversed_symbol):
            symbols_to_add = [reversed_symbol]

        # Add new symbols to watched currencies.
        # Skip initializing_symbol_prices_pairs as they already have been added
        # Skip exchange_config.traded_symbol_pairs as they are already added to regular price feeds
        new_symbols_to_add = [
            s
            for s in symbols_to_add
            if s not in self.initializing_symbol_prices_pairs
            and s not in self.portfolio_manager.exchange_manager.exchange_config.traded_symbol_pairs
        ]
        if new_symbols_to_add:
            self.logger.debug(f"Fetching price for {new_symbols_to_add} to compute all the "
                              f"currencies values and properly evaluate portfolio.")
            self._ask_ticker_data_for_currency(new_symbols_to_add)
            self.initializing_symbol_prices.add(currency)
            self.initializing_symbol_prices_pairs.update(new_symbols_to_add)

    def _ask_ticker_data_for_currency(self, symbols_to_add):
        """
        Synchronously call TICKER_CHANNEL producer to add a list of new symbols to its watch list
        :param symbols_to_add: the list of symbol to add to the TICKER_CHANNEL producer watch list
        """
        if self._bot_main_loop is asyncio.get_event_loop():
            asyncio.create_task(
                self.portfolio_manager.exchange_manager.exchange_config.add_watched_symbols(symbols_to_add)
            )
        else:
            asyncio_tools.run_coroutine_in_asyncio_loop(
                self.portfolio_manager.exchange_manager.exchange_config.add_watched_symbols(symbols_to_add),
                self._bot_main_loop
            )

    def _inform_no_matching_symbol(self, currency):
        """
        Log a missing currency pair to calculate the portfolio profitability
        :param currency: the concerned currency
        """
        # do not log warning in backtesting or tests
        if not self.portfolio_manager.exchange_manager.is_backtesting:
            self.logger.warning(f"No trading pair including {currency} and {self.portfolio_manager.reference_market} on"
                                f" {self.portfolio_manager.exchange_manager.exchange_name}. {currency} "
                                f"can't be valued for portfolio and profitability.")

    def convert_currency_value_using_last_prices(
        self, quantity, current_currency, target_currency, settlement_asset=None
    ):
        try:
            symbol = symbol_util.merge_currencies(
                current_currency, target_currency, settlement_asset=settlement_asset
            )
            if self._has_price_data(symbol):
                return quantity * self._get_last_price_data(symbol)
        except KeyError:
            pass
        try:
            reversed_symbol = symbol_util.merge_currencies(
                target_currency, current_currency, settlement_asset=settlement_asset
            )
            return quantity / self._get_last_price_data(reversed_symbol)
        except (KeyError, decimal.DivisionByZero):
            pass
        raise errors.MissingPriceDataError(
            f"no price data to evaluate {current_currency} price in {target_currency}"
        )

    def try_convert_currency_value_using_multiple_pairs(
            self, currency, target, quantity, base_bridge
    ) -> decimal.Decimal:
        # settlement_asset needs to be handled to add support for futures

        # try with two pairs
        # for example:
        # currency: ETH - ref market: USDT
        # ETH/USDT is not available. ETH/BTC and BTC/USDT are available though.
        # first convert ETH -> BTC and then BTC -> USDT
        #               | bridge part 1     | bridge part 2

        try:
            return self.convert_currency_value_from_saved_price_bridges(currency, target, quantity)
        except errors.MissingPriceDataError:
            if self.is_missing_price_bridge(currency, target):
                return None
            # try to find a bridge
        if len(base_bridge) > self.MAX_PRICE_BRIDGE_DEPTH:
            self._save_missing_price_bridge(currency, target)
            return None
        part_1_base = currency
        part_2_quote = target
        # look into available symbols to find pair bridges
        for bridge_part_1_symbol in self._get_priced_pairs():
            parsed_bridge_part_1_symbol = symbol_util.parse_symbol(bridge_part_1_symbol)
            if (parsed_bridge_part_1_symbol.base, parsed_bridge_part_1_symbol.quote) in base_bridge\
               or (parsed_bridge_part_1_symbol.quote, parsed_bridge_part_1_symbol.base) in base_bridge:
                # avoid looping in symbols
                continue
            # part 1: check if bridge_part_1_symbol can be used as part 1 of the bridge
            is_inverse_part_1 = False
            if part_1_base == parsed_bridge_part_1_symbol.quote:
                is_inverse_part_1 = True
            elif part_1_base != parsed_bridge_part_1_symbol.base:
                continue
            part_1_quote = parsed_bridge_part_1_symbol.quote
            if is_inverse_part_1:
                part_1_quote = parsed_bridge_part_1_symbol.base
                part_1_base = parsed_bridge_part_1_symbol.quote
            try:
                bridge_part_1_value = self.convert_currency_value_using_last_prices(
                    quantity, part_1_base, part_1_quote
                )
                # check that bridge_part_1_value is really set
                if not bridge_part_1_value:
                    continue
            except errors.MissingPriceDataError:
                # first pair might not be initialized or is not available at all
                # make sure it's not just initializing
                self._ensure_no_pending_symbol_price(part_1_base, part_1_quote)
                # try with other pairs
                continue
            # part 2: bridge part 1 is found and valued, try to get a compatible second part
            # case 1: 2 parts bridge
            try:
                bridge_part_2_value = self.convert_currency_value_using_last_prices(
                    constants.ONE,
                    part_1_quote,
                    part_2_quote,
                )
                if bridge_part_2_value:
                    self._remove_from_missing_currency_data(currency)
                    local_bridge = [(part_1_base, part_1_quote), (part_1_quote, part_2_quote)]
                    self._save_price_bridge(currency, target, local_bridge)
                    return bridge_part_1_value * bridge_part_2_value
            except errors.MissingPriceDataError:
                # conversion pairs might not be initialized or is not available at all
                self._ensure_no_pending_symbol_price(part_1_quote, part_2_quote)
                # otherwise continue with other pairs
            bridge = base_bridge + [(part_1_base, part_1_quote)]
            # case 2: X parts bridge
            nested_value = self.try_convert_currency_value_using_multiple_pairs(
                part_1_quote, target, constants.ONE, bridge
            )
            if nested_value:
                try:
                    extended_bridge = [(part_1_base, part_1_quote)] \
                        + self.get_saved_price_conversion_bridge(part_1_quote, target)
                    self._save_price_bridge(currency, target, extended_bridge)
                except KeyError:
                    # should not happen, however if it does, do not crash
                    pass
                return bridge_part_1_value * nested_value
        # no bridge found
        self._save_missing_price_bridge(currency, target)
        return None

    def _get_priced_pairs(self):
        for pair in self.last_prices_by_trading_pair:
            # first look into pairs with price
            yield pair
        for pair in self.portfolio_manager.exchange_manager.exchange_config.traded_symbol_pairs:
            # then into pairs from config
            if pair not in self.last_prices_by_trading_pair:
                yield pair
        if self.initializing_symbol_prices:
            for pair in self.initializing_symbol_prices_pairs:
                # finally into initializing pairs
                yield pair

    def _ensure_no_pending_symbol_price(self, base, quote):
        for symbol in (symbol_util.merge_currencies(base, quote), symbol_util.merge_currencies(quote, base)):
            if symbol in self.portfolio_manager.exchange_manager.exchange_config.traded_symbol_pairs \
                    or symbol in self.initializing_symbol_prices_pairs:
                raise errors.PendingPriceDataError

    def get_saved_price_conversion_bridge(self, currency, target) -> list:
        return self._price_bridge_by_symbol[symbol_util.merge_currencies(currency, target)]

    def _save_price_bridge(self, currency, target, bridge):
        self._price_bridge_by_symbol[symbol_util.merge_currencies(currency, target)] = bridge

    def convert_currency_value_from_saved_price_bridges(self, currency, target, quantity) -> decimal.Decimal:
        try:
            bridge = self._price_bridge_by_symbol[symbol_util.merge_currencies(currency, target)]
            converted_value = quantity
            for base, quote in bridge:
                converted_value = self.convert_currency_value_using_last_prices(converted_value, base, quote)
            return converted_value
        except KeyError as err:
            raise errors.MissingPriceDataError from err

    def reset_missing_price_bridges(self):
        self._missing_price_bridges = set()

    def _save_missing_price_bridge(self, base, quote):
        self._missing_price_bridges.add(symbol_util.merge_currencies(base, quote))

    def is_missing_price_bridge(self, base, quote):
        return symbol_util.merge_currencies(base, quote) in self._missing_price_bridges

    def _remove_from_missing_currency_data(self, currency):
        if currency in self.missing_currency_data_in_exchange:
            self.missing_currency_data_in_exchange.remove(currency)

    def _has_price_data(self, symbol):
        return self._get_last_price_data(symbol) is not constants.ZERO

    def _get_last_price_data(self, symbol):
        try:
            return self.last_prices_by_trading_pair[symbol]
        except KeyError:
            # a settlement asset or other symbol extra 
            # data might be different, try to ignore it
            to_find_symbol = symbol_util.parse_symbol(symbol)
            for symbol_key, last_prices in self.last_prices_by_trading_pair.items():
                if symbol_util.parse_symbol(symbol_key).is_same_base_and_quote(to_find_symbol):
                    return last_prices
        raise KeyError(symbol)

    def clear(self):
        self.portfolio_manager = None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

from mock import patch, Mock, AsyncMock
import pytest

import octobot_trading.personal_data
import octobot_trading.errors
import octobot_trading.enums as enums

from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import buy_limit_order, stop_loss_limit_order

pytestmark = pytest.mark.asyncio


async def test_constructor(buy_limit_order):
    # with normal order
    state = octobot_trading.personal_data.OrderState(buy_limit_order, True)
    assert state.order is buy_limit_order

    # with cleared order
    buy_limit_order.state = state
    buy_limit_order.clear()
    with pytest.raises(octobot_trading.errors.InvalidOrderState):
        octobot_trading.personal_data.OrderState(buy_limit_order, True)


async def test_update_calls_synchronize_when_pending_for_buy_limit_order(buy_limit_order):
    buy_limit_order.order_type = enums.TraderOrderType.BUY_LIMIT
    state = octobot_trading.personal_data.OrderState(buy_limit_order, True)
    if os.getenv('CYTHON_IGNORE'):
        return
    with patch.object(state, 'synchronize', new=AsyncMock()) as order_state_synchronize_mock, \
            patch.object(state, 'terminate', new=AsyncMock()) as order_state_terminate_mock:
        await state.update()
        order_state_synchronize_mock.assert_called_once()
        order_state_terminate_mock.assert_not_called()


async def test_update_calls_terminate_when_pending_for_stop_loss_limit_order(stop_loss_limit_order):
    stop_loss_limit_order.order_type = enums.TraderOrderType.STOP_LOSS_LIMIT
    state = octobot_trading.personal_data.OrderState(stop_loss_limit_order, True)
    if os.getenv('CYTHON_IGNORE'):
        return
    with patch.object(state, 'synchronize', new=AsyncMock()) as order_state_synchronize_mock, \
            patch.object(state, 'terminate', new=AsyncMock()) as order_state_terminate_mock:
        await state.update()
        order_state_synchronize_mock.assert_not_called()
        order_state_terminate_mock.assert_called_once()


async def test_update_calls_nothing_when_refreshing(buy_limit_order):
    buy_limit_order.order_type = enums.TraderOrderType.BUY_LIMIT
    state = octobot_trading.personal_data.OrderState(buy_limit_order, True)
    if os.getenv('CYTHON_IGNORE'):
        return
    with patch.object(state, 'synchronize', new=AsyncMock()) as order_state_synchronize_mock, \
            patch.object(state, 'terminate', new=AsyncMock()) as order_state_terminate_mock:
        state.state = enums.States.REFRESHING
        await state.update()
        order_state_synchronize_mock.assert_not_called()
        order_state_terminate_mock.assert_not_called()


async def test_is_created(buy_limit_order):
    buy_limit_order.order_type = enums.TraderOrderType.BUY_LIMIT
    state = octobot_trading.personal_data.OrderState(buy_limit_order, True)
    assert state.is_created() is True


async def test_replace_order(buy_limit_order, stop_loss_limit_order):
    buy_limit_order.order_type = enums.TraderOrderType.BUY_LIMIT
    state = octobot_trading.personal_data.OrderState(buy_limit_order, True)
    buy_limit_order.state = state
    assert state.order is buy_limit_order
    assert buy_limit_order.state is state
    await state.replace_order(stop_loss_limit_order)
    assert state.order is stop_loss_limit_order
    assert stop_loss_limit_order.state is state
    assert buy_limit_order.state is None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses


import octobot_commons.minimizable_dataclass as minimizable_dataclass


@dataclasses.dataclass
class ExchangeAuthDetails:
    api_key: str = ""
    api_secret: str = ""
    api_password: str = ""
    exchange_type: str = ""
    sandboxed: bool = False
    encrypted: bytes = ""


@dataclasses.dataclass
class ExchangeDetails:
    name: str


@dataclasses.dataclass
class MarketDetails:
    id: str
    symbol: str
    info: dict
    time_frame: str
    close: list[float]
    open: list[float]
    high: list[float]
    low: list[float]
    volume: list[float]
    time: list[float]


@dataclasses.dataclass
class OrdersDetails:
    open_orders: list[dict] = None
    missing_orders: list[dict] = None

    def __post_init__(self):
        if self.open_orders is None:
            self.open_orders = []
        if self.missing_orders is None:
            self.missing_orders = []


@dataclasses.dataclass
class PortfolioDetails:
    initial_value: float = 0
    content: dict = None
    asset_values: dict = None

    def __post_init__(self):
        if self.content is None:
            self.content = {}
        if self.asset_values is None:
            self.asset_values = {}


@dataclasses.dataclass
class ExchangeData(minimizable_dataclass.MinimizableDataclass):
    auth_details: ExchangeAuthDetails
    exchange_details: ExchangeDetails
    markets: list[MarketDetails] = None
    portfolio_details: PortfolioDetails = None
    orders_details: OrdersDetails = None
    trades: list[dict] = None

    # pylint: disable=E1134
    def __post_init__(self):
        if isinstance(self.auth_details, dict):
            self.auth_details = ExchangeAuthDetails(**self.auth_details)
        if isinstance(self.exchange_details, dict):
            self.exchange_details = ExchangeDetails(**self.exchange_details)
        if self.markets is None:
            self.markets = []
        elif self.markets and isinstance(self.markets[0], dict):
            self.markets = [MarketDetails(**market) for market in self.markets] if self.markets else []
        if not isinstance(self.portfolio_details, PortfolioDetails):
            self.portfolio_details = PortfolioDetails(**self.portfolio_details) if \
                self.portfolio_details else PortfolioDetails()
        if not isinstance(self.orders_details, OrdersDetails):
            self.orders_details = OrdersDetails(**self.orders_details) if self.orders_details else OrdersDetails()
        if self.trades is None:
            self.trades = []

    def get_price(self, symbol):
        for market in self.markets:
            if market.symbol == symbol:
                return market.close[-1]
        raise KeyError(symbol)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import async_channel.constants as constants
import octobot_trading.exchange_channel as exchanges_channel


class RecentTradeProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, symbol, recent_trades, replace_all=False):
        await self.perform(symbol, recent_trades, replace_all=replace_all)

    async def perform(self, symbol, recent_trades, replace_all=False):
        try:
            if self.channel.get_filtered_consumers(symbol=constants.CHANNEL_WILDCARD) or \
                    self.channel.get_filtered_consumers(symbol=symbol):
                recent_trades = self.channel.exchange_manager.get_symbol_data(symbol).handle_recent_trade_update(
                    recent_trades,
                    replace_all=replace_all)

                if recent_trades:
                    await self.send(cryptocurrency=self.channel.exchange_manager.exchange.
                                    get_pair_cryptocurrency(symbol),
                                    symbol=symbol,
                                    recent_trades=recent_trades)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, recent_trades):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "recent_trades": recent_trades
            })


class RecentTradeChannel(exchanges_channel.ExchangeChannel):
    FILTER_SIZE = 10
    PRODUCER_CLASS = RecentTradeProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer


class LiquidationsProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, symbol, liquidations):
        await self.perform(symbol, liquidations)

    async def perform(self, symbol, liquidations):
        try:
            if self.channel.get_filtered_consumers(symbol=constants.CHANNEL_WILDCARD) or \
                    self.channel.get_filtered_consumers(symbol=symbol):
                self.channel.exchange_manager.get_symbol_data(symbol).handle_liquidations(liquidations)
                await self.send(cryptocurrency=self.channel.exchange_manager.exchange.get_pair_cryptocurrency(symbol),
                                symbol=symbol,
                                liquidations=liquidations)
        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def send(self, cryptocurrency, symbol, liquidations):
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "liquidations": liquidations
            })


class LiquidationsChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = LiquidationsProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import sortedcontainers

import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.util as util
from octobot_trading.enums import ExchangeConstantsOrderBookInfoColumns as ECOBIC

ORDER_ID_NOT_FOUND = -1
INVALID_PARSED_VALUE = -1


class OrderBookManager(util.Initializable):
    def __init__(self):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.order_book_initialized = False
        self.asks = sortedcontainers.SortedDict()
        self.bids = sortedcontainers.SortedDict()
        self.timestamp = 0
        self.ask_quantity, self.ask_price, self.bid_quantity, self.bid_price = 0, 0, 0, 0

    async def initialize_impl(self):
        self.reset_order_book()

    def reset_order_book(self):
        self.order_book_initialized = False
        self.asks.clear()
        self.bids.clear()
        self.timestamp = 0
        self.ask_quantity, self.ask_price, self.bid_quantity, self.bid_price = 0, 0, 0, 0

    def order_book_ticker_update(self, ask_quantity, ask_price, bid_quantity, bid_price):
        self.ask_quantity, self.ask_price = ask_quantity, ask_price
        self.bid_quantity, self.bid_price = bid_quantity, bid_price

    def handle_new_book(self, orders):
        try:
            self.handle_new_books(asks=orders[ECOBIC.ASKS.value],
                                  bids=orders[ECOBIC.BIDS.value],
                                  timestamp=orders[ECOBIC.TIMESTAMP.value])
        except KeyError:
            self.logger.error("Failed to parse new order book")

    def handle_new_books(self, asks, bids, timestamp=None):
        self.reset_order_book()
        self.handle_book_adds(_convert_price_size_list_to_order(asks, enums.TradeOrderSide.SELL.value))
        self.handle_book_adds(_convert_price_size_list_to_order(bids, enums.TradeOrderSide.BUY.value))
        if timestamp:
            self.timestamp = timestamp
        self.order_book_initialized = True

    def handle_book_adds(self, orders):
        for order in orders:
            try:
                self._handle_book_add(order)
            except KeyError as e:
                self.logger.error(f"Error when adding order to order_book : {e}")

    def handle_book_deletes(self, orders):
        for order in orders:
            try:
                self._handle_book_delete(order)
            except KeyError as e:
                self.logger.error(f"Error when deleting order from order_book : {e}")

    def handle_book_updates(self, orders):
        for order in orders:
            try:
                self._handle_book_update(order)
            except KeyError as e:
                self.logger.error(f"Error when updating order in order_book : {e}")

    def _handle_book_add(self, order):
        # Add buy side orders
        if order[ECOBIC.SIDE.value] == enums.TradeOrderSide.BUY.value:
            bids = self.get_bids(order[ECOBIC.PRICE.value])
            if bids is None:
                bids = [order]
            else:
                bids.append(order)
            self._set_bids(order[ECOBIC.PRICE.value], bids)
            return

        # Add sell side orders
        asks = self.get_asks(order[ECOBIC.PRICE.value])
        if asks is None:
            asks = [order]
        else:
            asks.append(order)
        self._set_asks(order[ECOBIC.PRICE.value], asks)

    def _handle_book_delete(self, order):
        price = decimal.Decimal(order[ECOBIC.PRICE.value])

        # Delete buy side orders
        if order[ECOBIC.SIDE.value] == enums.TradeOrderSide.BUY.value:
            bids = self.get_bids(price)
            if bids is not None:
                bids = [bid_order for bid_order in bids if
                        bid_order[ECOBIC.ORDER_ID.value] != order[ECOBIC.ORDER_ID.value]]
                if len(bids) > 0:
                    self._set_bids(price, bids)
                else:
                    self._remove_bids(price)
            return

        # Delete sell side orders
        asks = self.get_asks(price)
        if asks is not None:
            asks = [ask_order for ask_order in asks if
                    ask_order[ECOBIC.ORDER_ID.value] != order[ECOBIC.ORDER_ID.value]]
            if len(asks) > 0:
                self._set_asks(price, asks)
            else:
                self._remove_asks(price)

    def _handle_book_update(self, order):
        size = decimal.Decimal(order.get(ECOBIC.SIZE.value, INVALID_PARSED_VALUE))
        price = decimal.Decimal(order[ECOBIC.PRICE.value])

        # Update buy side orders
        if order[ECOBIC.SIDE.value] == enums.TradeOrderSide.BUY.value:
            bids = self.get_bids(price)
            order_index = _order_id_index(order[ECOBIC.ORDER_ID.value], bids)
            if bids is None or order_index == ORDER_ID_NOT_FOUND:
                return
            if size != INVALID_PARSED_VALUE:
                bids[order_index][ECOBIC.SIZE.value] = size
            self._set_bids(price, bids)
            return

        # Update sell side orders
        asks = self.get_asks(price)
        order_index = _order_id_index(order[ECOBIC.ORDER_ID.value], asks)
        if asks is None or order_index == ORDER_ID_NOT_FOUND:
            return
        if size != INVALID_PARSED_VALUE:
            asks[order_index][ECOBIC.SIZE.value] = size
        self._set_asks(price, asks)

    def _set_asks(self, price, asks):
        self.asks[price] = asks

    def _set_bids(self, price, bids):
        self.bids[price] = bids

    def _remove_asks(self, price):
        del self.asks[price]

    def _remove_bids(self, price):
        del self.bids[price]

    def get_ask(self):
        return self.asks.peekitem(0)

    def get_bid(self):
        return self.bids.peekitem(-1)

    def get_asks(self, price):
        return self.asks.get(price, None)

    def get_bids(self, price):
        return self.bids.get(price, None)


def _order_id_index(order_id, order_list):
    """
    Return order id index in order list
    :param order_id: the order id to search
    :param order_list: the order list to check
    :return: the order id index in order list if found else ORDER_ID_NOT_FOUND
    """
    if order_list is not None:
        for index, order in enumerate(order_list):
            if order[ECOBIC.ORDER_ID.value] == order_id:
                return index
    return ORDER_ID_NOT_FOUND


def _convert_price_size_list_to_order(price_size_list, side):
    """
    Convert a [price, size] list to the book order format
    :param price_size_list: the list of [price, size]
    :param side: the order side
    :return: the converted list
    """
    return [
        _convert_price_size_to_order(price_size, side)
        for price_size in price_size_list
    ]


def _convert_price_size_to_order(price_size, side):
    """
    Convert a pair of price and size to an order book dict format
    :param price_size: the pair, size pair to convert
    :param side: the order side
    :return: the converted pair dict
    """
    return {
        ECOBIC.SIDE.value: side,
        ECOBIC.PRICE.value: price_size[0],
        ECOBIC.SIZE.value: price_size[1],
        ECOBIC.ORDER_ID.value: None
    }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.enums as commons_enums
import octobot_commons.databases as commons_databases


def set_plot_orders(ctx, value):
    ctx.plot_orders = value


async def disable_candles_plot(ctx, exchange_manager=None):
    storage = (exchange_manager or ctx.exchange_manager).storage_manager.candles_storage
    await storage.enable(False)
    await storage.clear_history()


async def clear_orders_cache(writer):
    await _clear_table(writer, commons_enums.DBTables.ORDERS.value)
    await writer.clear()


async def clear_symbol_plot_cache(writer):
    await _clear_table(writer, commons_enums.DBTables.CACHE_SOURCE.value)
    await writer.clear()


async def _clear_table(writer, table, flush=True):
    await writer.delete(table, None)
    if flush:
        await writer.flush()


def get_shared_element(key):
    return commons_databases.GlobalSharedMemoryStorage.instance()[key]


def set_shared_element(key, element):
    commons_databases.GlobalSharedMemoryStorage.instance()[key] = element



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.channels as channels

import octobot_backtesting.api as api
import octobot_backtesting.enums as backtesting_enums

import octobot_commons.channels_name as channels_name
import octobot_commons.enums as common_enums
import octobot_commons.errors as errors

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.exchange_data.recent_trades.channel.recent_trade_updater as recent_trade_updater
import octobot_trading.util as util


class RecentTradeUpdaterSimulator(recent_trade_updater.RecentTradeUpdater):
    SIMULATED_RECENT_TRADE_LIMIT = 2

    def __init__(self, channel, importer):
        super().__init__(channel)
        self.exchange_data_importer = importer
        self.exchange_name = self.channel.exchange_manager.exchange_name

        self.last_timestamp_pushed = 0
        self.last_timestamp_pushed_by_symbol = {}
        self.time_consumer = None
        # Only generate recent trades from the shortest handled time frame
        self.recent_trades_time_frame = self.channel.exchange_manager.exchange_config.get_shortest_time_frame().value

    async def start(self):
        await self.resume()

    async def handle_timestamp(self, timestamp, **kwargs):
        try:
            for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                recent_trades_data = (await self.exchange_data_importer.get_recent_trades_from_timestamps(
                    exchange_name=self.exchange_name,
                    symbol=pair,
                    inferior_timestamp=timestamp,
                    limit=1))[0]
                if recent_trades_data[0] > self.last_timestamp_pushed:
                    self.last_timestamp_pushed = recent_trades_data[0]
                    await self.push(pair, recent_trades_data[-1])
        except errors.DatabaseNotFoundError as e:
            self.logger.warning(f"Not enough data : {e} will use ohlcv data to simulate recent trades.")
            await self.pause()
            await self.stop()

    async def _recent_trades_from_ohlcv_callback(self, exchange: str, exchange_id: str,
                                                 cryptocurrency: str, symbol: str, time_frame, candle):
        if time_frame == self.recent_trades_time_frame:
            try:
                # Candles are pushed when completed therefore the current price is the candle's close price
                # However we can't only rely on close price to generate minimal recent trades.
                # Recent trades for this candle are between the next candle candle high price and the next candle
                # low price in order to get these prices, use exchange simulator's future candles filled by the
                # previous ohlcv updater cycle that also triggered this call
                future_candle = self.channel.exchange.get_current_future_candles()[symbol][time_frame]
                last_candle_timestamp = future_candle[common_enums.PriceIndexes.IND_PRICE_TIME.value]
                if last_candle_timestamp > self.last_timestamp_pushed_by_symbol[symbol]:
                    future_candle_low_price = future_candle[common_enums.PriceIndexes.IND_PRICE_LOW.value]
                    future_candle_high_price = future_candle[common_enums.PriceIndexes.IND_PRICE_HIGH.value]
                    recent_trades = [self._generate_recent_trade(last_candle_timestamp, future_candle_low_price),
                                     self._generate_recent_trade(last_candle_timestamp, future_candle_high_price)]
                    self.last_timestamp_pushed_by_symbol[symbol] = last_candle_timestamp
                    await self.push(symbol, recent_trades)
            except (KeyError, TypeError):
                # future candle not initialized or missing, should rarely happen: use received candle's close value
                if candle:
                    last_candle_timestamp = candle[common_enums.PriceIndexes.IND_PRICE_TIME.value]
                    if last_candle_timestamp > self.last_timestamp_pushed_by_symbol[symbol]:
                        last_candle_close_price = candle[common_enums.PriceIndexes.IND_PRICE_CLOSE.value]
                        recent_trades = [self._generate_recent_trade(last_candle_timestamp, last_candle_close_price)] \
                                        * self.SIMULATED_RECENT_TRADE_LIMIT
                        self.last_timestamp_pushed_by_symbol[symbol] = last_candle_timestamp
                        await self.push(symbol, recent_trades)

    @staticmethod
    def _generate_recent_trade(timestamp, price):
        return {
            enums.ExchangeConstantsOrderColumns.TIMESTAMP.value: timestamp,
            enums.ExchangeConstantsOrderColumns.PRICE.value: price
        }

    async def pause(self):
        if self.time_consumer is not None:
            await channels.get_chan(channels_name.OctoBotBacktestingChannelsName.TIME_CHANNEL.value). \
                remove_consumer(self.time_consumer)
        self.is_running = False

    async def stop(self):
        await util.stop_and_pause(self)

    async def resume(self):
        if not self.is_running:
            if self.time_consumer is None and not self.channel.is_paused:
                if backtesting_enums.ExchangeDataTables.RECENT_TRADES in \
                        api.get_available_data_types(self.exchange_data_importer):
                    self.time_consumer = await channels.get_chan(
                        channels_name.OctoBotBacktestingChannelsName.TIME_CHANNEL.value).new_consumer(
                        self.handle_timestamp)
                else:
                    await exchanges_channel.get_chan(constants.OHLCV_CHANNEL, self.channel.exchange_manager.id) \
                        .new_consumer(self._recent_trades_from_ohlcv_callback)
                    self.last_timestamp_pushed_by_symbol = {
                        symbol: 0
                        for symbol in self.channel.exchange_manager.exchange_config.traded_symbol_pairs
                    }
                self.is_running = True
